#include "record.h"
#include "ui_record.h"
#include <QTimer>
#include <QAudio>
#include <QMessageBox>

// this defines the somewhat arbitrary size of PCM buffers to send data to LAME encoder
#define PCMBUFSIZE  2048
short int Buffer_l[PCMBUFSIZE]; // buffer for left channel raw PCM data
short int Buffer_r[PCMBUFSIZE]; // buffer for right channel raw PCM data
unsigned int BufferCount;
unsigned char mp3buf[PCMBUFSIZE + (PCMBUFSIZE+3)/4 + 7200]; // this should be big enough for the MP3 output from LAME

Record::Record(QWidget *parent) :
    QDialog(parent),
    ui(new Ui::Record)
{
    QList<QAudioDeviceInfo> devices;
    QAudioFormat format;
    int i;

    ui->setupUi(this);
    mw=(MainWindow *)(this->parent());
    showLevel=FALSE;
    isRecording=FALSE;
    recordingTrack=0;
    id3Track=0;
    recordingdevice=NULL;
    recordingio=NULL;
    ui->label_Right->setVisible(mw->channels==2);
    ui->label_RightLevel->setVisible(mw->channels==2);
    ui->label_RightPeak->setVisible(mw->channels==2);
    ui->progressBarVURight->setVisible(mw->channels==2);
    QTimer *timer = new QTimer(this);
    connect(timer, SIGNAL(timeout()), this, SLOT(OneSec()));
    timer->start(1000); // start executing OneSec() every  second
    ui->progressBarPeakRight->setVisible(mw->channels==2);
    connect(ui->pushButtonExit,SIGNAL(clicked()),this,SLOT(close()));
    connect(ui->pushButtonStartStopRecording,SIGNAL(clicked()),this,SLOT(startstopRecording()));
    connect(ui->pushButtonLevel,SIGNAL(clicked()),this,SLOT(toggleShowLevel()));
    devices=QAudioDeviceInfo::availableDevices(QAudio::AudioInput);
    for (i=0; i<devices.size(); i++) {
        if (devices[i].deviceName()==mw->source) {
            break;
        }
    }
    if (i<devices.size()) {
        recordingdeviceinfo=devices[i];
        format.setByteOrder(QAudioFormat::LittleEndian);
        format.setCodec("audio/pcm");
        format.setSampleType(QAudioFormat::SignedInt);
        format.setSampleSize(16);
        format.setChannelCount(mw->channels);
        format.setSampleRate(mw->rate);
        recordingdevice=new QAudioInput(recordingdeviceinfo,format,this);
        openMP3file();
        recordingio=recordingdevice->start();
        connect(recordingio,SIGNAL(readyRead()),this,SLOT(getaudio()));
        startvumeter();
    } else {
        QMessageBox::warning(this,"Error","Audio input device not found");
    }
}

Record::~Record()
{
    if (isRecording) {
        isRecording=FALSE;
        closeMP3file(); // close any recording in progress (do not open another file)
    }
    recordingdevice->stop();
    delete recordingdevice;
    delete ui;
}

void Record::toggleShowLevel(void)
{
    showLevel=!showLevel;
    if (!showLevel) {
        ui->pushButtonLevel->setText("Level");
    }
}

void Record::OneSec(void)
{
    if (isRecording) {
        recLength++;
        updateStatus();
    }
}

// update status waiting to record or recording (with length)
void Record::updateStatus(void)
{
    int hour,minute,second;

    if (isRecording) {
        second=recLength%60;
        minute=(recLength/60)%60;
        hour=recLength/3600;
        ui->labelStatus->setText(QString("Recording track %1 %2:%3:%4")
                                 .arg(recordingTrack,3,10,QChar('0'))
                                 .arg(hour)
                                 .arg(minute,2,10,QChar('0'))
                                 .arg(second,2,10,QChar('0')));
    } else {
        ui->labelStatus->setText(QString("Waiting to record track %1").arg(recordingTrack,3,10,QChar('0')));
    }
}

// initialize all data used by the VU and peak meters
// including filtering time constant and sample window sizes
// VU meter updates every 50 ms (20 times/sec)
// peak meter updates every 500 ms (2 times/second)
// VU low pass filter time constant is .065 seconds, yeilding a 0% to 99% rise time of 300 ms
void Record::startvumeter(void)
{
    kynhpleft=kynhpright=0;
    xnm1left=xnm1right=0;
    khpm1=(double)mw->rate/(2.0*3.14159*20.0); // k-1 for highpass cutoff frequency of 20 Hz
    khp=khpm1+1; // k for highpass cutoff frequency of 20 Hz
    kynleft=kynright=0;
    kvu=((double)mw->rate)/16*.065;
    vusamples=mw->rate/16/20; // filter each 16 samples, update meter 20 times/second
    peakleft=peakright=0;
    peaksamples=mw->rate/2; // update peak meter 2 times/second
    soundcount=0;
}

void Record::getaudio(void)
{
    QByteArray data;

    data=recordingio->readAll();
    processAudio((qint16 *)data.data(),data.count()/2);
}

// samples is the number of 16 bit words of data
// if channels==2, the left and right channels each contain half that many samples (interleaved L/R/L/R)
// Vu meter requires computing an average (i.e. low pass filter) of the absolute value of the input
// compute a lowpass filter based on Yn = Yn-1 + (Xn - Yn-1)/K, where K = samplerate*timeconstant
// for integer math, compute K*Yn = K*Yn-1 + Xn - K*Yn-1/K
// for efficiency, only compute the filter every 16 samples based on average of absolute values of 16 input samples
// also keep track of peak signal level for periodic updating of peak level meters
// and send the audio data to the lame encoder (with stereo/mono conversion as necessary)
// Caution: if the input data has any DC offset, the measured signal level (Vu meter) will have a positive offset
// This makes it difficult to detect low level input signals for sound avtivated recording
// To avoid problems, Vu meter measures the input data after passing it through a high pass filter to remove DC
// A simple first order high pass filter is computed as Yn = a*(Yn-1 + Xn - Xn-1), where a = 1/(2*pi*fc/samplerate + 1)
// Note that "a" will be just slightly less that 1 for a low cutoff frequency fc.
// To compute the filter with all integer math, I transform using a = 1 - 1/k, or k = 1/(1-a)
// then compute k*Yn instead of just Yn
// actual computation becomes k*Yn = k*Yn-1 - k*Yn-1/k + (k-1)*(Xn - Xn-1)
// I use a high pass cutoff frequency of 20 Hz
void Record::processAudio(const qint16 *data, int samples)
{
    static int count16=0;   // counter for averaging 16 audio samples before computing VU lowpass filter
    static int countfilters=0;  // counter for how many time to process VU lowpass filter before updating the VU level meters
    static int leftsum=0,rightsum=0;    // sums for averaging 16 audio samples
    int newkynleft,newkynright; // new VU filter data to display on VU meters
    static int countpeak=0; // count of how many audio samples to process before updating the peak level meters
    int newpeakleft,newpeakright;   // new peak data to diaplay on the peak level meters
    int i;
    qint16 datal,datar;
    int absdata;

    if (mw->channels==1) { // process mono audio data
        newkynleft=-1;
        newpeakleft=-1;
        for (i=0; i<samples; i++) {
            datal=*data++;  // get left channel data (only channel for mono)
            if (isRecording) {  // record the data if recording is active
                lameencode(datal,datal); // duplatate the mono data on left and right channels in case MP3 output is stereo
            }
            kynhpleft=kynhpleft-kynhpleft/khp+khpm1*(datal-xnm1left); // compute highpass filter to remove DC offset
            xnm1left=datal; // save xn-1 left channel (previous xn value needed for high pass filter)
            absdata=abs(kynhpleft/khp); // use the highpass filter output for Vu and peak meters
            if (absdata>peakleft) { // if new peak
                peakleft=absdata;
            }
            leftsum=leftsum+absdata; // average the data for level meter
            if (++countpeak>=peaksamples) { // if time to update peak meter
                countpeak=0;
                if (peakleft>newpeakleft) {
                    newpeakleft=peakleft;
                }
                peakleft=0;
            }
            if (++count16>=16) { // if time to do low pass filtering for level meter
                count16=0;
                kynleft=kynleft+leftsum/16-kynleft/kvu;
                leftsum=0; // reset sum to start averaging the next 16 samples
                if (++countfilters>=vusamples) { // if time to update the level meter
                    countfilters=0;
                    // update the vu meter
                    newkynleft=kynleft*2; // arbitrary scale factor because average meter tends to read low
                    if (mw->soundactivated) {
                        autostartstop(newkynleft);
                    }
                }
            }
        }
        if (newkynleft>=0) { // if new data for level meter
            ui->progressBarVULeft->setValue(ui->progressBarVULeft->maximum()-newkynleft/kvu);
            ui->progressBarVULeft->repaint();   // must repaint every time so meter updates continuously instead of in large steps
            if (showLevel) {
                ui->pushButtonLevel->setText(QString("%1\%").arg((200*newkynleft/kvu)/32768));
            }
        }
        if (newpeakleft>=0) { // if new data for peak meter
            ui->progressBarPeakLeft->setValue(ui->progressBarPeakLeft->maximum()-newpeakleft);
            ui->progressBarPeakLeft->repaint();   // must repaint every time so meter updates continuously instead of in large steps
        }
    } else {
        if (mw->channels==2) { // process stereo audio data
            samples=samples/2; // only half the samples for each channel
            newkynleft=newkynright=-1;
            newpeakleft=newpeakright=-1;
            for (i=0; i<samples; i++) {
                datal=*data++;
                datar=*data++;
                kynhpleft=kynhpleft-kynhpleft/khp+khpm1*(datal-xnm1left); // compute highpass filter to remove DC offset
                xnm1left=datal; // save xn-1 left channel (previous xn value needed for high pass filter)
                kynhpright=kynhpright-kynhpright/khp+khpm1*(datar-xnm1right); // compute highpass filter to remove DC offset
                xnm1right=datar; // save xn-1 right channel (previous xn value needed for high pass filter)
                if (isRecording) {  // record the data if recording is active
                    if (MP3IsStereo) {
                        lameencode(datal,datar); // stereo uses both channels
                    } else {
                        lameencode((datal+datar)/2,0); // mono uses mix (average) of both channels
                    }
                }
                absdata=abs(kynhpleft/khp); // use the highpass filter output for Vu and peak meters
                if (absdata>peakleft) { // if new peak left
                    peakleft=absdata;
                }
                leftsum=leftsum+absdata; // average data for level meter left
                absdata=abs(kynhpright/khp); // use the highpass filter output for Vu and peak meters
                if (absdata>peakright) { // if new peak right
                    peakright=absdata;
                }
                rightsum=rightsum+absdata; // average data for level meter right
                if (++countpeak>=peaksamples) { // if time to update peak meters
                    countpeak=0;
                    if (peakleft>newpeakleft) {
                        newpeakleft=peakleft;
                    }
                    if (peakright>newpeakright) {
                        newpeakright=peakright;
                    }
                    peakleft=0;
                    peakright=0;
                }
                if (++count16>=16) { // if time to do low pass filtering for level meters
                    count16=0;
                    kynleft=kynleft+leftsum/16-kynleft/kvu;
                    leftsum=0;
                    kynright=kynright+rightsum/16-kynright/kvu;
                    rightsum=0;
                    if (++countfilters>=vusamples) { // if time to update the level meters
                        countfilters=0;
                        // update the vu meters
                        newkynleft=kynleft*2; // arbitrary scale factor because average meter tends to read low
                        newkynright=kynright*2; // arbitrary scale factor because average meter tends to read low
                        if (mw->soundactivated) {
                            autostartstop(newkynleft>newkynright ? newkynleft : newkynright); // in stereo I use max of right/left levels
                        }
                    }
                }
            }
            if (newkynleft>=0) { // if new data for level meters
                ui->progressBarVULeft->setValue(ui->progressBarVULeft->maximum()-newkynleft/kvu);
                ui->progressBarVULeft->repaint();   // must repaint every time so meter updates continuously instead of in large steps
                ui->progressBarVURight->setValue(ui->progressBarVURight->maximum()-newkynright/kvu);
                ui->progressBarVURight->repaint();   // must repaint every time so meter updates continuously instead of in large steps
                if (showLevel) {
                    if (newkynright>newkynleft) {
                        newkynleft=newkynright;
                    }
                    ui->pushButtonLevel->setText(QString("%1\%").arg((200*newkynleft/kvu)/32768));
                }
            }
            if (newpeakleft>=0) { // if new data for peak meters
                ui->progressBarPeakLeft->setValue(ui->progressBarPeakLeft->maximum()-newpeakleft);
                ui->progressBarPeakLeft->repaint();   // must repaint every time so meter updates continuously instead of in large steps
                ui->progressBarPeakRight->setValue(ui->progressBarPeakRight->maximum()-newpeakright);
                ui->progressBarPeakRight->repaint();   // must repaint every time so meter updates continuously instead of in large steps
            }
        }
    }
}

// called after each 50 ms of sound is processed for the VU level meter
// compares levels against start/stop level thresholds and times the start/stop times
// sets the Record pushbutton to checked state when recording should start
// clears the Record pushbutton to unchecked state when recording should stop
void Record::autostartstop(int level)
{
    level=(200*level/kvu)/32768; // I multiply by 200 instead of 100 so start/stop thresholds of 0-100 represent 0-50% on the VU meter
    if (isRecording) { // if recording is in progress
        if (level<mw->stoplevel) {
            if ((++soundcount)*5>=mw->stoptime) { // stop recording now
                soundcount=0;
                stopRecording();
            }
        } else {
            soundcount=0;
        }
    } else { // not recording
        if (level>=mw->startlevel) {
            if ((++soundcount)*5>=mw->starttime) { // start recording now
                soundcount=0;
                startRecording();
            }
        } else {
            soundcount=0;
        }
    }
}

void Record::startstopRecording(void)
{
    if (isRecording) {
        stopRecording();
    } else {
        startRecording();
    }
}

// everything is already to go, so just set the flag to begin recording
// and update status message on screen
void Record::startRecording(void)
{
    isRecording=TRUE;
    recLength=0;
    updateStatus();
    ui->pushButtonStartStopRecording->setText("&Stop Recording");
}

// must finish up and close the MP3 stream and output file, and prepage a new MP3 stream and MP3 output file
void Record::stopRecording(void)
{
    isRecording=FALSE;
    closeMP3file();
    openMP3file();
}

// prepare an MP3 stream for encoding and an MP3 file for output
// this is done before starting the recording so the recording can start faster
// and must be done when recording is stopped to prepare for the next recording
void Record::openMP3file(void)
{
    static int MP3Rate[] = {0,8000,11025,12000,16000,22050,24000,32000,44100,48000}; // these are the MP3 sample rates
    static int MP3Bitrate[] = {8,16,24,32,40,48,56,64,80,96,112,128,144,160,192,224,256,320}; // these are the MP3 bitrates
    static MPEG_mode MP3Channels[] = {MONO,STEREO,JOINT_STEREO};
    static vbr_mode MP3Encoding[] = {vbr_off,vbr_abr,vbr_mtrh};
    static int MP3Quality[] = {7,5,2}; // quality settings for low (good but fast), medium (very good), high (excellent but slower)

    lgf=mw->lameinit(); // create a lame encoder instance
    mw->lamesetinsamplerate(lgf,mw->rate);  // input sample rate
    mw->lamesetnumchannels(lgf,mw->MP3Channels==0 ? 1 : 2);   // input channels (I cheat and tell encoder it matches output channels)
    mw->lamesetoutsamplerate(lgf,MP3Rate[mw->MP3Frequency]); // output sample rate
    mw->lamesetbwritevbrtag(lgf,0); // never write vbr tag
    mw->lamesetquality(lgf,MP3Quality[mw->MP3Quality]);  // quality average for now
    mw->lamesetmode(lgf,MP3Channels[mw->MP3Channels]);  // output channel mode
    mw->lamesetbrate(lgf,MP3Bitrate[mw->MP3Bitrate]);   // output bitrate
    mw->lamesetcopyright(lgf,0);    // no copyright
    mw->lamesetoriginal(lgf,1); // yes original
    mw->lameseterrorprotection(lgf,0);  // no crc
    mw->lamesetvbr(lgf,MP3Encoding[mw->MP3Encoding]);   // select CBR/ABR/VBR
    if (mw->lameinitparams(lgf)!=0) {    // set all the parameters
        QMessageBox::warning(this,"Error","Unable to open MP3 stream\nCheck your settings.");
    }
    do {
        MP3File.setFileName(QString("%1/%2%3.mp3").arg(mw->savefolder).arg(mw->getFilename()).arg(++recordingTrack,3,10,QChar('0')));
    } while (MP3File.exists()); // keep trying next track number until the file does not exist
    MP3File.open(QIODevice::WriteOnly);
    BufferCount=0;
    MP3IsStereo=(mw->MP3Channels!=0);
    updateStatus();
    ui->pushButtonStartStopRecording->setText("&Start Recording");
}

// Call lameencode for each and every sample to be encoded
// For mono output only left channel will be used
// Samples are put in Buffer_l and Buffer_r until buffer is full
// When buffer is full, data is encoded by lame and written to file
// and then the buffer is reset to empty
// Note, any partial buffer will be encoded later when the MP3 stream is closed)
void Record::lameencode(short int data_l, short int data_r)
{
    int MP3Size;

    Buffer_l[BufferCount]=data_l;
    Buffer_r[BufferCount]=data_r;
    BufferCount++;
    if (BufferCount >= PCMBUFSIZE) {
        MP3Size=mw->lameencodebuffer(lgf,Buffer_l,Buffer_r,BufferCount,mp3buf,sizeof(mp3buf));
        if (MP3Size>0) {
            MP3File.write((const char *)mp3buf,MP3Size);
        }
        BufferCount=0; // empty the buffers
    }
}

void Record::closeMP3file(void)
{
    int MP3Size;

    // finish up the MP3 file
    if (BufferCount > 0) { // buffer not empty, encode it
        MP3Size=mw->lameencodebuffer(lgf,Buffer_l,Buffer_r,BufferCount,mp3buf,sizeof(mp3buf));
        if (MP3Size>0) {
            MP3File.write((const char *)mp3buf,MP3Size);
        }
    }
    MP3Size=mw->lameencodeflush(lgf,mp3buf,sizeof(mp3buf));
    if (MP3Size>0) {
        MP3File.write((const char *)mp3buf,MP3Size);
    }
    mw->lameclose(lgf);
    writetags();
    MP3File.close();
}

void Record::writetags(void)
{
    struct id3v1 {
        char header[3];
        char title[30];
        char artist[30];
        char album[30];
        char year[4];
        char comment[28];
        char zero;
        char track;
        char genre;
    } tags;
    if (MP3File.pos()>0) { // only write tags if file is not empty
        if (artist!="" || album!="" || comment!="" || year!= "") { // only write tags if at least one tag has data
            memset(&tags,0,sizeof(tags));
            memcpy(tags.header,"TAG",sizeof(tags.header));
            memcpy(tags.artist,artist.toAscii(),artist.toAscii().size());
            memcpy(tags.album,album.toAscii(),album.toAscii().size());
            memcpy(tags.comment,comment.toAscii(),comment.toAscii().size());
            memcpy(tags.year,year.toAscii(),year.toAscii().size());
            tags.track=++id3Track;
            tags.genre=0xff; // (none)
            MP3File.write((const char *)&tags,sizeof(tags));
        }
    }
}
