#include "settings.h"
#include "ui_settings.h"
#include <QFileDialog>
#include <QAudio>

/* Note:
When the setting dialog in opened, the constructor executes first.
This populates the source list of soundcards.
Later setSource is calles, selecting the previously selected sound card (if it is found in the list).
Following that, setRate and setChannels are called, selecting the previously selected values (if found).
*/
Settings::Settings(QWidget *parent) :
    QDialog(parent),
    ui(new Ui::Settings)
{
    ui->setupUi(this);
    connect(ui->pushButtonSaveFolder,SIGNAL(clicked()),this,SLOT(doBrowseSaveFolder()));
    connect(ui->comboBoxSource,SIGNAL(currentIndexChanged(int)),this,SLOT(deviceSelected(int)));
    enumerateSoundCards(); // populate the list of available sound cards
}

Settings::~Settings()
{
    delete ui;
}

void Settings::setSource(QString source)
{
    int i;
    int sourceindex;

    sourceindex=-1;
    for (i=0; i<devices.size(); i++) {
        if (devices[i].deviceName()==source) {
            sourceindex=i;
            break;
        }
    }
    if (sourceindex>=0) {
        ui->comboBoxSource->setCurrentIndex(sourceindex); // select the current sound device
    }
}

void Settings::setRate(int rate)
{
    int i;
    int rateindex;

    rateindex=-1;
    for (i=0; i<ui->comboBoxRate->count(); i++) {
        if (ui->comboBoxRate->itemData(i).toInt()==rate) {
            rateindex=i;
        }
    }
    if (rateindex>=0) {
        ui->comboBoxRate->setCurrentIndex(rateindex);
    }
}

void Settings::setChannels(int channels)
{
    int i;
    int channelsindex;

    channelsindex=-1;
    for (i=0; i < ui->comboBoxChannels->count(); i++) {
        if (ui->comboBoxChannels->itemData(i).toInt()==channels) {
            channelsindex=i;
        }
    }
    if (channelsindex>=0) {
        ui->comboBoxChannels->setCurrentIndex(channelsindex);
    }
}

void Settings::setSoundActivated(bool soundactivated)
{
    ui->checkBoxSoundActivated->setChecked(soundactivated);
}

void Settings::setStartLevel(int startlevel)
{
    ui->spinBoxStartLevel->setValue(startlevel);
}

void Settings::setStartTime(int starttime)
{
    ui->doubleSpinBoxStartTime->setValue(((double)starttime)/100);
}

void Settings::setStopLevel(int stoplevel)
{
    ui->spinBoxStopLevel->setValue(stoplevel);
}

void Settings::setStopTime(int stoptime)
{
    ui->doubleSpinBoxStopTime->setValue(((double)stoptime)/100);
}

void Settings::setSaveFolder(QString savefolder)
{
    ui->lineEditSaveFolder->setText(savefolder);
}

void Settings::setMP3Channels(int index)
{
    ui->comboBoxMP3Channels->setCurrentIndex(index);
}

void Settings::setMP3Frequency(int index)
{
    ui->comboBoxMP3Frequency->setCurrentIndex(index);
}

void Settings::setMP3Bitrate(int index)
{
    ui->comboBoxMP3Bitrate->setCurrentIndex(index);
}

void Settings::setMP3Encoding(int index)
{
    ui->comboBoxMP3Encoding->setCurrentIndex(index);
}

void Settings::setMP3Quality(int index)
{
    ui->comboBoxMP3Quality->setCurrentIndex(index);
}

QString Settings::getSource(void)
{
    int index;

    index=ui->comboBoxSource->currentIndex();
    if (index>=0) {
        return devices[index].deviceName();
    }
    return ""; // no devices!
}

int Settings::getRate(void)
{
    int index;

    index=ui->comboBoxRate->currentIndex();
    if (index>=0) {
        return ui->comboBoxRate->itemData(index).toInt();
    }
    return 0;
}

int Settings::getChannels(void)
{
    int index;

    index=ui->comboBoxChannels->currentIndex();
    if (index>=0) {
        return ui->comboBoxChannels->itemData(index).toInt();
    }
    return 0;
}

bool Settings::getSoundActivated(void)
{
    return ui->checkBoxSoundActivated->isChecked();
}

int Settings::getStartLevel(void)
{
    return ui->spinBoxStartLevel->value();
}

int Settings::getStartTime(void)
{
    return (int)(ui->doubleSpinBoxStartTime->value()*100);
}

int Settings::getStopLevel(void)
{
    return ui->spinBoxStopLevel->value();
}

int Settings::getStopTime(void)
{
    return (int)(ui->doubleSpinBoxStopTime->value()*100+.5); // add .5 to avoid problems with floating point rounding errors
}

QString Settings::getSaveFolder(void)
{
    return ui->lineEditSaveFolder->text();
}

int Settings::getMP3Channels(void)
{
    return ui->comboBoxMP3Channels->currentIndex();
}

int Settings::getMP3Frequency(void)
{
    return ui->comboBoxMP3Frequency->currentIndex();
}

int Settings::getMP3Bitrate(void)
{
    return ui->comboBoxMP3Bitrate->currentIndex();
}

int Settings::getMP3Encoding(void)
{
    return ui->comboBoxMP3Encoding->currentIndex();
}

int Settings::getMP3Quality(void)
{
    return ui->comboBoxMP3Quality->currentIndex();
}

void Settings::doBrowseSaveFolder(void)
{
    QString newfolder=QFileDialog::getExistingDirectory(this,"Save Folder",ui->lineEditSaveFolder->text());
    if (newfolder!="") {
        ui->lineEditSaveFolder->setText(newfolder);
    }
}
void Settings::enumerateSoundCards(void)
{
    int i;

    devices=QAudioDeviceInfo::availableDevices(QAudio::AudioInput);
    ui->comboBoxSource->clear();
    for (i=0; i<devices.size(); i++) {
        ui->comboBoxSource->addItem(devices[i].deviceName());
    }
}

// a sound card has been selected in the source list, time to populate the channels and sample rate lists
void Settings::deviceSelected(int index)
{
    QList<int> values;
    int i;
    int currentindex;
    int currentchannels,currentrate;

    currentchannels=-1;
    currentindex=ui->comboBoxChannels->currentIndex();
    if (currentindex>=0) {
        currentchannels=ui->comboBoxChannels->itemData(currentindex).toInt();
    }
    currentrate=-1;
    currentindex=ui->comboBoxRate->currentIndex();
    if (currentindex>=0) {
        currentrate=ui->comboBoxRate->itemData(currentindex).toInt();
    }
    ui->comboBoxChannels->clear();
    ui->comboBoxRate->clear();
    if (index>=0) {
        currentindex=-1;
        values=devices[index].supportedChannels();
        for (i=0; i<values.size(); i++) {
            ui->comboBoxChannels->addItem(QString::number(values[i]),values[i]);
            if (values[i]==currentchannels) {
                currentindex=i;
            }
        }
        if (currentindex>=0) {
            ui->comboBoxChannels->setCurrentIndex(currentindex);
        }
        currentindex=-1;
        values=devices[index].supportedSampleRates();
        for (i=0; i<values.size(); i++) {
            ui->comboBoxRate->addItem(QString::number(values[i]),values[i]);
            if (values[i]==currentrate) {
                currentindex=i;
            }
        }
        if (currentindex>=0) {
            ui->comboBoxRate->setCurrentIndex(currentindex);
        }
    }
}
